/**
 * main.c
 *
 * GenerateUHG is a generator for graphs with a given number k of hamiltonian cycles 
 * (which is especially efficient for small values of k).
 *
 * The latest version of GenerateUHG can be found here:
 * http://caagt.ugent.be/uhg/
 * 
 * Author: Barbara Meersman
 * In collaboration with Jan Goedgebeur and Carol Zamfirescu
 *
 * -----------------------------------
 * 
 * Copyright (c) 2017-2018 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * Compile with:
 *
 * cc -DWORDSIZE=32 -DMAXN=WORDSIZE -mpopcnt -O3 planarity.c general.c canonical_construction_path.c nautyW1.a main.c -o generateUHG
 *
 */

#include "general.h"

static void
perform_wordsize_checks() {
    /* Checks to test if the WORDSIZE and sizeof setwords is valid */
    if (WORDSIZE != 32 && WORDSIZE != 64) {
        fprintf(stderr, "Error: invalid value for wordsize: %d\n", WORDSIZE);
        fprintf(stderr, "Valid values are 32 and 64\n");
        exit(1);
    }

    if (MAXN != WORDSIZE) {
        fprintf(stderr, "Error: MAXN is not equal to WORDSIZE: %d vs %d\n", MAXN, WORDSIZE);
        fprintf(stderr, "Please compile with option -DMAXN=WORDSIZE\n");
        exit(1);
    }

    if (MAXM != 1) {
        fprintf(stderr, "Error: expected MAXM 1 instead of %d\n", MAXM);
        exit(1);
    }

    if (WORDSIZE == 32) {
        fprintf(stderr, "32 bit mode\n");
        if (sizeof (unsigned int) < 4) {
            fprintf(stderr, "Error: unsigned ints should be at least 32 bit -- sorry.\n");
            exit(1);
        }
        if (sizeof (setword) < 4) {
            fprintf(stderr, "Error: this version relies on 32 bit setwords -- sorry.\n");
            exit(1);
        }
        if (sizeof (setword) != 4) {
            fprintf(stderr, "Error: expected 32 bit setwords\n");
            exit(1);
        }
    } else {
        fprintf(stderr, "64 bit mode\n");
        if (sizeof (setword) != 8) {
            fprintf(stderr, "Error: this version relies on 64 bit setwords -- sorry.\n");
            exit(1);
        }
        if (sizeof (unsigned long long int) != 8) {
            fprintf(stderr, "Error: this version relies on 64 bit unsigned long long ints -- sorry.\n");
            exit(1);
        }
    }
}

int main(int argc, char *argv[]) {

    perform_wordsize_checks();
    nauty_check(WORDSIZE, MAXM, MAXN, NAUTYVERSIONID);

    float savetime_start;
    float savetime_end;

    if (argc < 2) {
        fprintf(stderr, "Error: invalid number of arguments. At least the number of vertices must be specified.\n");
        fprintf(stderr, "Usage: %s <number_of_vertices> [options]\n", argv[0]);
        fprintf(stderr, "Execute '%s -h' for more extensive help.\n", argv[0]);
        exit(1);
    }

    if (argc == 2 && argv[1][0] == '-' && argv[1][1] == 'h') {
        print_help(argv[0]);
        exit(1);
    }

    sscanf(argv[1], "%d", &number_of_vertices);

    if (number_of_vertices < 1) {
        fprintf(stderr, "Error: the number of vertices must be at least 1\n");
        exit(1);
    }
    
    if (number_of_vertices > MAXN) {
        fprintf(stderr, "Error: too many vertices (max %d)\n", MAXN);
        exit(1);
    }


    triangle_free = 0;
    square_free = 0;
    planar = 0;
    maxdeg = number_of_vertices;
    max_number_of_nodes_maxdeg = number_of_vertices + 1;
    max_number_of_cycles = 1;
    mindeg_2 = 0;
    no_n_minus_one = 0;
    number_of_times_max_edges = 0;
    output_exact = 1;
    int print_edgestats = 0;

    int i;
    for (i = 2; i < argc; i++) {
        switch (argv[i][0]) {
            case 'H':
            {
                max_number_of_cycles = atoi(argv[i] + 1);
                if (max_number_of_cycles < 0) {
                    fprintf(stderr, "Error: the number of hamiltonian cycles must be at least 0\n");
                    exit(1);
                }
                if (max_number_of_cycles > MAXN) {
                    fprintf(stderr, "Error: the number of hamiltonian cycles can be at most %d\n", MAXN);
                    exit(1);
                }
                break;
            }
            case 'D':
            {
                maxdeg = atoi(argv[i] + 1);
                if (maxdeg < 0) {
                    fprintf(stderr, "Error: maximum degree must be a natural number\n");
                    exit(1);
                }
                fprintf(stderr, "Info: only outputting graphs with maxdeg at most %d\n", maxdeg);
                break;
            }
            case 'T':
            {
                triangle_free = 1;
                fprintf(stderr, "Info: searching for triangle-free graphs\n");
                break;
            }
            case 'F':
            {
                square_free = 1;
                fprintf(stderr, "Info: searching for square-free graphs\n");
                break;
            }
            case 'M':
            {
                if(maxdeg == number_of_vertices) {
                    fprintf(stderr, "Error: maxdeg (option D<x>) must be set first!\n");
                    exit(1);
                }
                max_number_of_nodes_maxdeg = atoi(argv[i] + 1);
                fprintf(stderr, "Info: searching for graphs with max %d nodes of maxdeg %d\n", max_number_of_nodes_maxdeg, maxdeg);
                break;
            }
            case 'P':
            {
                planar = 1;
                fprintf(stderr, "Info: searching for planar graphs\n");
                break;
            }
            case 'C':
            {
                no_n_minus_one = 1;
                fprintf(stderr, "Info: only outputting graphs without n-1 cycles\n");
                break;
            }
            case 'o':
            {
                output_exact = 0;
                break;
            }
            case 'e':
            {
                if (strcmp(argv[i], "edgestats") == 0) {
                    print_edgestats = 1;
                    break;
                } else {
                    fprintf(stderr, "Error: invalid option\n");
                    exit(1);
                }
            }            
            case 'S':
            {
                modulo = 1;
                i++;
                rest = atoi(argv[i]);
                i++;
                mod = atoi(argv[i]);
                if (rest >= mod) {
                    fprintf(stderr, "Error: rest (%d) must be smaller than mod (%d).\n", rest, mod);
                    exit(1);
                }
                break;
            }
            default:
            {
                fprintf(stderr, "Error: invalid option: %s\n", argv[i]);
                fprintf(stderr, "Execute '%s -h' for a list of possible options.\n", argv[0]);
                exit(1);
            }
        }
    }
    
    if(!output_exact && max_number_of_cycles < 2)  {
        fprintf(stderr, "Error: option 'o' only works if the number of hamiltonian cycles is at least 2\n");
        exit(1);
    }
    
    if(!output_exact)
        fprintf(stderr, "Info: outputting hamiltonian graphs with at most %d hamiltonian cycles\n", max_number_of_cycles);
    else
        fprintf(stderr, "Info: outputting graphs with exactly %d hamiltonian cycle(s)\n", max_number_of_cycles);
    
    if (maxdeg == 2) {
        current_number_of_nodes_maxdeg = number_of_vertices;
    } else {
        current_number_of_nodes_maxdeg = 0;
    }

    if (modulo) {
        splitlevel_counter = 0;
        splitlevel = 4;
        if (max_number_of_cycles == 0) {
            //splitlevel = number_of_vertices - 1;
            splitlevel = number_of_vertices + 3;
        }
        if(triangle_free || square_free || planar){
        	splitlevel++;
        }    

        fprintf(stderr, "Info: splitlevel is %d\n", splitlevel);
    }

    number_of_nauty_calls = 0;
    number_of_edges_before_nauty_call = 0;
    number_of_non_canonical_edges = 0;
    number_of_graphs_tested = 0;
    num_graphs_tested_for_planarity = 0;

    //For the timing
    struct tms TMS;

    for (i = 0; i < argc; i++)
        fprintf(stderr, "%s ", argv[i]);
    fprintf(stderr, "\n");


    times(&TMS);
    savetime_start = TMS.tms_utime;

    if (max_number_of_cycles == 0 || number_of_vertices > 1) {
        generate();
    }

    times(&TMS);
    savetime_end = TMS.tms_utime;

    //	fprintf(stderr, "Number of graphs tested: %d\n", number_of_graphs_tested);
    //	fprintf(stderr, "Number of nauty calls on not canonical graphs: %f\n", number_of_non_canonical_edges/number_of_graphs_generated);
    //	fprintf(stderr, "Average number of edges before nauty call: %f\n", number_of_edges_before_nauty_call);
/*
    for (i = 0; i < 6; i++) {
        fprintf(stderr, "Percentage of graphs accepted/rejected by criterium %d: %llu (%.1f%%)\n", i, number_of_graphs_by_criterium[i], (double) 100 * number_of_graphs_by_criterium[i] / number_of_graphs_tested);
    }
*/
    //fprintf(stderr, "Percentage of graphs rejected because of hamiltonian cycle: %llu (%.1f%%)\n", reject_by_ham_cycle, (double) 100 * reject_by_ham_cycle / number_of_graphs_tested);
    //fprintf(stderr, "Number of nauty calls: %llu (%.1f%%)\n", number_of_nauty_calls, (double) 100 * number_of_nauty_calls / number_of_graphs_tested);

    //	fprintf(stderr, "Number of graphs tested for canonicity: %d\n", number_of_graphs_tested);
    //	fprintf(stderr, "Max number of nodes of maximum degree %d\n", max_number_of_nodes_maxdeg);
/*
    if (planar) {
        //fprintf(stderr, "Number of graphs tested for planarity: %llu\n", num_graphs_tested_for_planarity);
        fprintf(stderr, "Number of graphs rejected for planarity: %llu (%.1f%%)\n", num_graphs_rejected_for_planarity, (double) 100 * num_graphs_rejected_for_planarity / num_graphs_tested_for_planarity);
    }
*/
    if (print_edgestats) {
        unsigned long long int total_num_graphs = 0;
        for (i = 0; i < number_of_vertices * (number_of_vertices - 1); i++) {
            if (number_of_graphs_with_given_num_edges[i] > 0) {
                fprintf(stderr, "Number of graphs with %d edges: %llu\n", i, number_of_graphs_with_given_num_edges[i]);
                total_num_graphs += number_of_graphs_with_given_num_edges[i];
            }
        }
        fprintf(stderr, "Total num graphs: %llu\n", total_num_graphs);
    }
    fprintf(stderr, "Number of graphs generated: %llu\n", number_of_graphs_written);
    //	fprintf(stderr, "Number of times maximum edges is reached %d\n", 	number_of_times_max_edges);
    fprintf(stderr, "Total CPU time: %.1f seconds.\n", (double) ((savetime_end - savetime_start) / time_factor));

    return 0;
}